import asyncio
import functools
import logging
import sys
import traceback
from dataclasses import dataclass
from hashlib import md5
from mimetypes import guess_type

from hrepr.resource import JSExpression


class StarbearHandler(logging.StreamHandler):
    def format(self, record):
        def _brack(s):
            return f"[\033[36m{s}\033[0m]" if s else ""

        process = getattr(record, "proc", None)
        user = getattr(record, "user", None)
        tb = getattr(record, "traceback", None)
        colors = {
            "INFO": "32",
            "WARNING": "33",
            "ERROR": "31",
        }
        color = colors.get(record.levelname, "95")
        prefix = f"\033[{color}m{record.levelname}\033[0m:   {_brack(record.name)}{_brack(process)}{_brack(user)}"
        msg = record.msg
        if tb:
            msg += "\n" + traceback.format_exc()
        if "\n" in msg:
            lines = f"\n\033[{color}m>\033[0m ".join(msg.split("\n"))
            return f"{prefix} {lines}"
        else:
            return f"{prefix} {msg}"


logger = logging.getLogger("starbear")
logger.setLevel(level=logging.INFO)
logger.addHandler(StarbearHandler(sys.stderr))

ABSENT = object()


def keyword_decorator(deco):
    """Wrap a decorator to optionally takes keyword arguments."""

    @functools.wraps(deco)
    def new_deco(fn=ABSENT, *args, **kwargs):
        if callable(fn):
            return deco(fn, *args, **kwargs)

        if fn is not ABSENT:
            args = (fn, *args)

        @functools.wraps(deco)
        def newer_deco(fn):
            return deco(fn, *args, **kwargs)

        return newer_deco

    return new_deco


class Queue(asyncio.Queue):
    def putleft(self, entry):
        self._queue.appendleft(entry)
        self._unfinished_tasks += 1
        self._finished.clear()
        self._wakeup_next(self._getters)

    def tag(self, tag):
        return ClientWrap(self, partial=[tag], pack=True)

    def wrap(self, **options):
        return ClientWrap(self, **options)

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await self.get()


class FeedbackQueue(Queue):
    pass


class VirtualFile:
    def __init__(self, content, type=None, name=None):
        if type is None:
            if name is not None:
                type, _ = guess_type(url=name)

        self.type = type
        self.content = content
        self.name = md5(content.encode("utf8")).hexdigest()
        if name is not None:
            self.name += f"/{name}"


@dataclass
class ClientWrap:
    DEFAULTS = {
        "debounce": 0,
        "extract": None,
        "form": False,
        "pack": False,
        "partial": None,
        "toggles": None,
        "pre": None,
        "post": None,
    }

    def __init__(self, func, **options):
        if any((key := k) not in self.DEFAULTS for k in options):
            raise TypeError(f"Invalid argument to ClientWrap: '{key}'")

        if isinstance(func, ClientWrap):
            inherit = func.options
            func = func.func
        else:
            inherit = self.DEFAULTS

        options = {**inherit, **options}

        partial = options["partial"]
        if partial is None or isinstance(partial, (list, tuple)):
            options["partial"] = partial
        else:
            options["partial"] = [partial]

        options["id"] = id(func)

        for x in ["pre", "post"]:
            result = options.get(x, None)
            if result is not None:
                if not isinstance(result, (list, tuple)):
                    result = [result]
                result = [
                    JSExpression(f"function (result) {{ {p} }}")
                    if isinstance(p, str)
                    else p
                    for p in result
                ]
            options[x] = result

        self.func = func
        self.options = options

    def wrap(self, **options):
        return type(self)(self, **options)

    def __aiter__(self):
        return self.func

    def __js_embed__(self, representer):
        fn = representer.js_embed(self.func)
        options = representer.js_embed(self.options)
        return f"$$BEAR_WRAP({fn}, {options})"

    def __attr_embed__(self, representer, attr):
        return f"$$BEAR_EVENT.call(this, {representer.js_embed(self)})"



import asyncio as aio

from hrepr import H, Tag
from hrepr.hgen import ResourceDeduplicator
from hrepr.resource import Resource


class Page:
    def __init__(
        self,
        iq,
        oq,
        representer,
        selector=None,
        query_params={},
        session={},
        track_history=True,
        sent_resources=None,
        app=None,
        loop=None,
    ):
        self.iq = iq
        self.oq = oq
        self.selector = selector
        self.query_params = query_params
        self.session = session
        self.representer = representer
        self.track_history = track_history
        self.sent_resources = sent_resources or ResourceDeduplicator()
        self.tasks = set()
        self.app = app
        self.loop = loop or aio.get_running_loop()
        self.js = JavaScriptOperation(self, [])
        self.window = JavaScriptOperation(self, [], False)

    def __getitem__(self, selector):
        def _map_selector(x):
            if isinstance(x, Tag):
                tid = x.attributes.get("id", None)
                if not tid:
                    raise Exception("Cannot locate element because it has no id.")
                return f"#{tid}"
            elif isinstance(x, str):
                return x
            else:
                raise TypeError("Only str or Tag can be used as a selector")

        if not isinstance(selector, tuple):
            selector = (selector,)
        selector = " ".join(map(_map_selector, selector))

        if self.selector is not None:
            selector = f"{self.selector} {selector}"

        return self.page_select(selector)

    def page_select(self, selector):
        return type(self)(
            iq=self.iq,
            oq=self.oq,
            selector=selector,
            representer=self.representer,
            track_history=self.track_history,
            sent_resources=self.sent_resources,
        )

    def with_history(self, track_history=True):
        if self.track_history == track_history:
            return self
        else:
            return type(self)(
                iq=self.iq,
                oq=self.oq,
                selector=self.selector,
                representer=self.representer,
                track_history=track_history,
                sent_resources=self.sent_resources,
            )

    def without_history(self):
        return self.with_history(False)

    def _push(self, coro):
        if aio._get_running_loop() is None:
            aio._set_running_loop(self.loop)
        task = aio.create_task(coro)
        self.tasks.add(task)
        task.add_done_callback(self.tasks.discard)

    async def sync(self):
        for task in list(self.tasks):
            await task

    def _to_element(self, x):
        if isinstance(x, str):
            return H.span(x)
        else:
            return self.representer.hrepr(x)

    async def _put(self, element, method, history=None, send_resources=False):
        if history is None:
            history = self.track_history
        sel = self.selector or "body"
        element = element(hx_swap_oob=f"{method}:{sel}")
        parts, extra, resources = self.representer.generate(
            element, filter_resources=self.sent_resources if send_resources else None
        )
        if not resources.empty():
            await self.page_select("head")._put(H.div(resources), "beforeend")
        to_send = str(parts)
        if extra:
            to_send += str(
                H.div(extra, style="display:none", hx_swap_oob=f"beforeend:{sel}")
            )
        return await self.oq.put((to_send, history))

    async def put(self, element, method, history=None):
        return await self._put(element, method, history=history, send_resources=True)

    def print(self, *elements, method="beforeend"):
        for element in elements:
            element = H.div(self._to_element(element))
            self._push(self.put(element, method))

    def print_html(self, html):
        self._push(self._put(H.div(H.raw(html)), "beforeend"))

    def set(self, element):
        element = H.div(self._to_element(element))
        self._push(self.put(element, "innerHTML"))

    def replace(self, element):
        element = self._to_element(element)
        self._push(self.put(element, "outerHTML"))

    def clear(self):
        self._push(self.put(H.span(), "innerHTML"))

    def delete(self):
        self._push(self.put(H.span(), "outerHTML"))

    def do(self, js):
        self.page_select("body").without_history().print(
            H.script(
                call_template.format(
                    selector=Resource(self.selector),
                    extractor=f"function () {{ {js} }}",
                    future="null",
                )
            ),
        )

    def toggle(self, toggle, value=None):
        return self.window["$$BEAR_TOGGLE"](self, toggle, value)

    def __js_embed__(self, representer):
        if self.selector is None:
            raise Exception(
                "Cannot send page reference to JS because it has no selector."
            )
        return f"document.querySelector('{self.selector}')"

    async def recv(self):
        return await self.iq.get()


call_template = "$$BEAR_CB({selector}, {extractor}, {future});"


def _extractor(sequence):
    result = "x"
    for entry in sequence:
        if isinstance(entry, str):
            result = f"{result}.{entry}"
        elif isinstance(entry, (list, tuple)):
            args = ",".join([str(Resource(x)) for x in entry])
            result = f"{result}({args})"
        else:
            raise TypeError()
    return f"(x => {result})"


class JavaScriptOperation:
    def __init__(self, element, sequence, selector=None):
        self.__element = element
        self.__sequence = sequence
        self.__future = aio.Future()
        self.__selector = (
            False if selector is False else (selector or self.__element.selector)
        )

    def __getattr__(self, attr):
        return type(self)(self.__element, [*self.__sequence, attr], self.__selector)

    __getitem__ = __getattr__

    def __call__(self, *args):
        return type(self)(self.__element, [*self.__sequence, args], self.__selector)

    def __await__(self):
        self.__element.page_select("body").without_history().print(
            H.script(
                call_template.format(
                    selector=Resource(self.__selector or None),
                    extractor=_extractor(self.__sequence),
                    future=Resource(self.__future),
                )
            ),
        )
        return iter(self.__future)



import asyncio as aio
import base64
import inspect
import json
import traceback
from functools import cached_property, wraps
from itertools import count
from pathlib import Path
from uuid import uuid4 as uuid

from hrepr import Tag
from starlette.exceptions import HTTPException
from starlette.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    Response,
)
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.websockets import WebSocket, WebSocketDisconnect

from .page import Page
from .repr import Representer
from .templating import template
from .utils import Queue, keyword_decorator, logger
from .wrap import with_error_display

here = Path(__file__).parent

_count = count()

construct = {}


def register_constructor(key):
    def deco(fn):
        construct[key] = fn
        return fn

    return deco


@register_constructor("HTMLElement")
def _(page, selector):
    return page[selector]


@register_constructor("Promise")
def _(page, id):
    async def resolve(value):
        await getattr(page.js, "$$BEAR_RESOLVE_LOCAL_PROMISE")(id, value)

    return resolve


def routeinfo(params="", path=None, root=False, cls=Route, **kw):
    def deco(method):
        assert method.__name__.startswith("route_")
        name = method.__name__.removeprefix("route_")
        method.routeinfo = {
            "cls": cls,
            "name": name,
            "path": path or f"/!/{name}",
            "root": root,
            "params": params,
            "keywords": kw,
        }
        return method

    return deco


class BasicBear:
    def __init__(self, fn):
        self.fn = fn
        self.__doc__ = getattr(fn, "__doc__", None)
        self.appid = next(_count)
        self._json_decoder = json.JSONDecoder(object_hook=self.object_hook)
        self.router = None
        self.route = None
        self.representer = None

    ###########
    # Methods #
    ###########

    def path_for(self, name, **kwargs):
        return self.router.url_path_for(self._mangle(name), **kwargs)

    def object_hook(self, dct):
        return dct

    async def json(self, request):
        body = await request.body()
        if isinstance(body, bytes):
            body = body.decode(encoding="utf8")
        return self._json_decoder.decode(body)

    ####################
    # Route generation #
    ####################

    def ensure_representer(self, request):
        router = request.scope["router"]
        if self.router is None:
            self.router = router
        else:
            assert self.router is router
        if self.representer is None:
            self.route = self.path_for("main").rstrip("/")
            self.representer = Representer(self.route)

    def wrap_route(self, method):
        @wraps(method)
        async def wrapped(request):
            self.ensure_representer(request)
            return await method(request)

        return wrapped

    def _mangle(self, name):
        return f"app{self.appid}_{name}"

    def _autoroutes(self):
        routes = []
        for method_name in dir(self):
            if method_name.startswith("route_"):
                method = getattr(self, method_name)
                routeinfo = method.routeinfo

                wmeth = self.wrap_route(method)
                mname = self._mangle(routeinfo["name"])
                route = routeinfo["cls"](
                    routeinfo["path"] + routeinfo["params"],
                    wmeth,
                    name=mname,
                    **routeinfo["keywords"],
                )
                routes.append(route)

                if routeinfo["root"]:
                    route = routeinfo["cls"](
                        "/",
                        wmeth,
                        name=mname,
                        **routeinfo["keywords"],
                    )
                    routes.append(route)
        return routes

    def routes(self):
        return self._autoroutes()

    ##############################
    # For standalone application #
    ##############################

    @cached_property
    def _mnt(self):
        return Mount("/", routes=self.routes())

    async def __call__(self, scope, receive, send):
        await self._mnt.handle(scope, receive, send)

    ################
    # Basic routes #
    ################

    @routeinfo("/{method:int}", methods=["GET", "POST"])
    async def route_method(self, request):
        method_id = request.path_params["method"]
        method = self.representer.callback_registry.resolve(method_id)
        try:
            args = await self.json(request)
        except json.JSONDecodeError:
            args = [await request.body()]
        result = method(*args, **request.query_params)
        if inspect.iscoroutine(result):
            result = await result
        if isinstance(result, Tag):
            return HTMLResponse(self.representer(result))
        else:
            return JSONResponse(result)

    @routeinfo(methods=["POST"])
    async def route_post(self, request):
        data = await self.json(request)
        if "error" in data:
            self.representer.future_registry.reject(
                fid=data["reqid"],
                error=data["error"],
            )
        else:
            self.representer.future_registry.resolve(
                fid=data["reqid"],
                value=data.get("value", None),
            )
        return JSONResponse({"status": "ok"})

    @routeinfo(methods=["POST"])
    async def route_queue(self, request):
        data = await self.json(request)
        self.representer.queue_registry.put(
            qid=data["reqid"],
            value=data["value"],
        )
        return JSONResponse({"status": "ok"})

    @routeinfo("/{path:path}")
    async def route_file(self, request):
        pth = self.representer.file_registry.get_file_from_url(
            request.path_params["path"]
        )
        if pth is None:
            raise HTTPException(
                status_code=404, detail="File not found or not available."
            )
        return FileResponse(pth, headers={"Cache-Control": "no-cache"})

    @routeinfo("/{path:path}")
    async def route_vfile(self, request):
        vf = self.representer.vfile_registry.get(request.path_params["path"])
        return Response(content=vf.content, media_type=vf.type)


class LoneBear(BasicBear):
    @routeinfo(path="/!", root=True)
    async def route_main(self, request):
        response = await self.fn(request)
        if isinstance(response, Tag):
            if response.name != "html":
                response = template(
                    here / "bare-template.html",
                    body=response,
                    route=self.route,
                    _asset=lambda name: here / name,
                )
            html = self.representer.generate_string(response)
            return HTMLResponse(html)
        elif isinstance(response, dict):
            return JSONResponse(response)
        else:
            return PlainTextResponse(str(response))


class Cub:
    def __init__(self, mother, process, query_params={}, session={}):
        self.mother = mother
        self.fn = mother.fn
        self.process = process
        self.query_params = query_params
        self.session = session
        self.route = self.mother.path_for("main", process=self.process).rstrip("/")
        self.representer = Representer(self.route)
        self.iq = Queue()
        self.oq = Queue()
        self.history = []
        self.reset = False
        self.ws = None
        self.page = Page(
            self.iq,
            self.oq,
            representer=self.representer,
            query_params=query_params,
            session=session,
            app=mother.app,
        )
        self.coro = aio.create_task(self.run())
        self._sd_coro = None
        self.log("info", "Created process")

    def log(self, level, msg, **extra):
        try:
            user = self.session.get("user", {}).get("email", None)
        except:
            logger.error("Could not get user")
        getattr(logger, level)(msg, extra={"proc": self.process, "user": user, **extra})

    def schedule_selfdestruct(self):
        async def sd():
            await aio.sleep(self.mother.process_timeout)
            del self.mother.cubs[self.process]
            self.coro.cancel()
            self.log("info", "Destroyed process")

        if self.mother.process_timeout is not None:
            self._sd_coro = aio.create_task(sd())

    def unschedule_selfdestruct(self):
        if self._sd_coro:
            self._sd_coro.cancel()
            self._sd_coro = None

    async def run(self):
        try:
            await self.fn(self.page)
            await self.page.sync()
        except Exception as exc:
            self.log("error", str(exc), traceback=traceback)
        finally:
            self.log("info", "Finished process")

    def _object_hook(self, dct):
        if "%" in dct:
            args = dict(dct)
            args.pop("%")
            return construct[dct["%"]](self.page, **args)
        else:
            return dct

    async def json(self, request):
        body = await request.body()
        return json.loads(body, object_hook=self._object_hook)

    async def route_main(self, request):
        self.unschedule_selfdestruct()
        node = template(
            here / "base-template.html",
            route=self.route,
            _asset=lambda name: here / name,
        )
        self.reset = True
        return HTMLResponse(
            self.representer.generate_string(node),
            headers={
                "Cache-Control": "no-cache, no-store, must-revalidate",
                "Pragma": "no-cache",
                "Expires": "0",
            },
        )

    async def route_socket(self, ws):
        self.unschedule_selfdestruct()

        async def recv():
            while True:
                try:
                    data = await ws.receive_json()
                    self.iq.put_nowait(data)
                except WebSocketDisconnect:
                    break

        async def send():
            while True:
                txt, in_history = await self.oq.get()
                try:
                    await ws.send_text(txt)
                    if in_history:
                        self.history.append(txt)
                except RuntimeError:
                    # Put the unsent element back into the queue
                    self.oq.putleft((txt, in_history))
                    break

        if self.ws:
            try:
                await self.ws.close()
            except RuntimeError:
                pass

        await ws.accept()
        self.ws = ws

        if self.reset:
            for entry in self.history:
                await ws.send_text(entry)
            self.reset = False

        await aio.wait(
            [aio.create_task(recv()), aio.create_task(send())],
            return_when=aio.FIRST_COMPLETED,
        )
        self.schedule_selfdestruct()

    async def route_method(self, request):
        method_id = request.path_params["method"]
        method = self.representer.callback_registry.resolve(method_id)
        try:
            args = await self.json(request)
        except json.JSONDecodeError:
            args = [await request.body()]
        result = method(*args, **request.query_params)
        if inspect.iscoroutine(result):
            result = await result
        if isinstance(result, Tag):
            return HTMLResponse(self.representer(result))
        else:
            return JSONResponse(result)

    async def route_post(self, request):
        data = await self.json(request)
        if "error" in data:
            self.representer.future_registry.reject(
                fid=data["reqid"],
                error=data["error"],
            )
        else:
            self.representer.future_registry.resolve(
                fid=data["reqid"],
                value=data.get("value", None),
            )
        return JSONResponse({"status": "ok"})

    async def route_queue(self, request):
        data = await self.json(request)
        self.representer.queue_registry.put(
            qid=data["reqid"],
            value=data["value"],
        )
        return JSONResponse({"status": "ok"})

    async def route_file(self, request):
        pth = self.representer.file_registry.get_file_from_url(
            request.path_params["path"]
        )
        if pth is None:
            raise HTTPException(
                status_code=404, detail="File not found or not available."
            )
        return FileResponse(pth, headers={"Cache-Control": "no-cache"})

    async def route_vfile(self, request):
        vf = self.representer.vfile_registry.get(request.path_params["path"])
        return Response(content=vf.content, media_type=vf.type)


def get_process_from_request(request):
    process_base = request.path_params.get("process", None)
    if process_base is None:
        process_base = base64.urlsafe_b64encode(uuid().bytes).decode("utf8").strip("=")
    return process_base


@keyword_decorator
def forward_cub(fn, ensure=False):
    @wraps(fn)
    async def fwd(self, request):
        self._ensure_router(request)
        process = get_process_from_request(request)
        cub = self._get(process, query_params=request.query_params, ensure=ensure)
        if cub is None:
            logger.warning(f"Trying to access missing process: {process}")
            if isinstance(request, WebSocket):
                await request.accept()
                await request.close(code=3002, reason="Missing application")
            else:
                return JSONResponse({"missing": process}, status_code=404)
        else:
            return await fn(self, request, cub)

    return fwd


class MotherBear:
    def __init__(self, fn, process_timeout=60, hide_processes=True):
        self.fn = fn
        self.__doc__ = getattr(fn, "__doc__", None)
        self.router = None
        self.process_timeout = process_timeout
        self.hide_processes = hide_processes
        self.cubs = {}
        self.appid = next(_count)

    def _get(self, proc, query_params={}, session={}, ensure=False):
        if proc not in self.cubs:
            if ensure:
                self.cubs[proc] = Cub(
                    self, proc, query_params=query_params, session=session
                )
            else:
                return None
        return self.cubs[proc]

    def path_for(self, name, **kwargs):
        return self.router.url_path_for(self._mangle(name), **kwargs)

    def _ensure_router(self, request):
        router = request.scope["router"]
        if self.router is None:
            self.router = router
            self.app = request.scope.get("app", None)
        else:
            assert self.router is router

    async def route_dispatch(self, request):
        self._ensure_router(request)
        process = get_process_from_request(request)
        main_path = self.path_for("main", process=process)
        if self.hide_processes:
            try:
                session = request.session
            except AssertionError:
                session = {}
            return await self._get(
                process, query_params=request.query_params, session=session, ensure=True
            ).route_main(request)
        else:
            url = main_path
            if request.query_params:
                url = f"{url}?{request.query_params}"
            return RedirectResponse(url=url)

    @forward_cub(ensure=True)
    async def route_main(self, request, cub):
        return await cub.route_main(request)

    @forward_cub
    async def route_socket(self, ws, cub):
        return await cub.route_socket(ws)

    @forward_cub
    async def route_method(self, request, cub):
        return await cub.route_method(request)

    @forward_cub
    async def route_file(self, request, cub):
        return await cub.route_file(request)

    @forward_cub
    async def route_vfile(self, request, cub):
        return await cub.route_vfile(request)

    @forward_cub
    async def route_post(self, request, cub):
        return await cub.route_post(request)

    @forward_cub
    async def route_queue(self, request, cub):
        return await cub.route_queue(request)

    def _mangle(self, name):
        return f"app{self.appid}_{name}"

    def _make_route(self, name, path, cls=Route, **kwargs):
        return cls(
            path,
            getattr(self, f"route_{name}"),
            name=self._mangle(name),
            **kwargs,
        )

    def routes(self):
        return [
            self._make_route("dispatch", "/"),
            self._make_route("main", "/!{process:str}/"),
            self._make_route(
                "method",
                "/!{process:str}/method/{method:int}",
                methods=["GET", "POST"],
            ),
            self._make_route("file", "/!{process:str}/file/{path:path}"),
            self._make_route("vfile", "/!{process:str}/vfile/{path:path}"),
            self._make_route("post", "/!{process:str}/post", methods=["POST"]),
            self._make_route("queue", "/!{process:str}/queue", methods=["POST"]),
            self._make_route("socket", "/!{process:str}/socket", cls=WebSocketRoute),
        ]

    @cached_property
    def _mnt(self):
        return Mount("/", routes=self.routes())

    async def __call__(self, scope, receive, send):
        await self._mnt.handle(scope, receive, send)


@keyword_decorator
def bear(fn, display_errors=True, **kwargs):
    if display_errors:
        fn = with_error_display(fn)
    return MotherBear(fn, **kwargs)


def simplebear(fn, **kwargs):
    return LoneBear(fn, **kwargs)



import traceback
from functools import wraps

from hrepr import H


def with_error_display(app):
    @wraps(app)
    async def wrapped_app(page):
        page["head"].print(
            H.style(
                """
        .STARBEAR__error:empty {
            display: none;
        }
        .STARBEAR__error {
            position: absolute;
            right: 0;
            top: 0;
            border: 1px solid red;
            padding: 3px;
            max-height: 300px;
            overflow: scroll;
            white-space: pre;
            color: black;
            background: white;
        }
        """
            )
        )
        page.print(err_div := H.div["STARBEAR__error"]().autoid())
        try:
            await app(page)
        except Exception as exc:
            page[err_div].set(
                H.div(
                    H.b("An error occurred. You may need to refresh the page.\n\n"),
                    traceback.format_exc(),
                )
            )
            raise

    return wrapped_app
